from typing import Dict, Optional
from uuid import uuid4

class Document:
    @classmethod
    def deserialize(cls, doc):
        return cls(doc.get('id', None), doc['externalId'], doc['contents'], doc.get('metadata', {}))

    def __init__(self, id, external_id, contents, metadata={}):
        self.id: str = id
        self.external_id: str = external_id
        self.contents: str = contents
        self.metadata: Dict = metadata

    def serialize(self):
        return {
            'id': self.id,
            'external_id': self.external_id,
            'contents': self.contents,
            'metadata': self.metadata
        }



from typing import List, Dict, Optional
from uuid import uuid4


class Annotation:
    def __init__(self, cls, selected_text, notes=None, raw=None):
        self.cls = cls
        self.selected_text = selected_text
        self.notes = notes
        self.raw = raw

    @classmethod
    def create_from_labeler(cls, project, obj):
        annotation_class = 'unknown'
        for fc in project.fragment_classes:
            if fc['id'] == obj['categoryId']:
                annotation_class = fc['name']

        return cls(annotation_class, obj['selectedText'], raw=obj)

    @classmethod
    def deserialize(cls, obj):
        return cls(obj['annotationCategory'], obj['selectedText'])

    def serialize(self, format='df'):
        if format == 'df':
            return {
                'class': self.cls,
                'selected_text': self.selected_text,
                'notes': self.notes
            }
        elif format == 'api':
            return self.raw
        else:
            raise Exception('Unsupported serialization format.')



class Label:
    @classmethod
    def deserialize_from_task(cls, task, response):
        annotations = [Annotation.deserialize(a) for a in response['contents']['annotations']]
        return cls(
            task.document.id,
            task.document.external_id,
            response['contents']['categoryName'],
            response['contents']['categoryId'],
            annotations,
            id=response['id']
        )

    @classmethod
    def deserialize(cls, label):
        annotations = [Annotation.deserialize(a) for a in label['annotations']]
        return cls(
            label['documentId'], # TODO-1: replace with id
            label['documentId'],
            label['documentCategory'],
            None, # class_id
            annotations=annotations
            # TODO-2: we need to pull id so that it isn't auto-generated
        )

    @classmethod
    def create_from_labeler(cls, document, class_name, class_id, project, annotations=None):
        if not annotations:
            annotations = []
        annotations = [Annotation.create_from_labeler(project, a) for a in annotations]
        return cls(document['id'], document['external_id'], class_name, class_id, annotations)

    def __init__(self, document_id, document_external_id, class_name, class_id, annotations=[], id=uuid4()):
        self.id: str = id
        self.document_id: str = document_id
        self.document_external_id: str = document_external_id
        self.cls: str = class_name
        self.cls_id: str = class_id

        if not annotations:
            annotations = []
        self.annotations: List[Annotation] = annotations # TODO-2: convert away from dict to Annotation class
        # TODO-3: creator information and other metadata, coming from API

    def serialize(self, project=None, format='api'):
        serialized = {
            'annotations': [a.serialize(format=format) for a in self.annotations]
        }

        if format == 'api':
            serialized['document_id'] = self.document_id
            serialized['class'] = self.cls
            serialized['class_id'] = self.cls_id
        elif format == 'df':
            serialized['document_id'] = self.document_external_id
            serialized['label'] = self.cls
        else:
            raise Exception('Unsupported serialization format.')

        return serialized



from typing import List, Optional, Dict

from .label import Label
from .document import Document

class Task:
    def __init__(self, id, is_complete, document):
        self.id = id
        self.is_complete = is_complete
        self.document = Document.deserialize(document)
        self.labels = []

    @classmethod
    def deserialize(cls, response):
        task = cls(response['id'], response['isComplete'], response['document'])
        for label_obj in response['labels']:
            label = Label.deserialize_from_task(task, label_obj)
            task.labels.append(label)
        return task



from typing import List, Optional, Dict
from uuid import uuid4, UUID
import pandas as pd # type: ignore

from .utils import chunker
from .labeler import Labeler
from .document import Document
from .task import Task


MAX_FILE_UPLOAD_BATCH_SIZE = 50


class Cycle:
    def __init__(self, project, cycle_data: Dict = {}): # TODO-3: fix project type, can't import due to circular import issue
        if cycle_data.get('id', None):
            self.id: str = cycle_data.get('id')
            self._is_new = False
        else:
            self.id: str = Cycle.generate_id()
            self._is_new = True

        self.project = project

        self.tasks: List[Task] = []
        self.documents: List[Document] = []

        self.deserialize_tasks(cycle_data.get('tasks', []))

    def __repr__(self):
        completed_tasks = [task for task in self.tasks if task.is_complete]
        incomplete_tasks = [task for task in self.tasks if not task.is_complete]
        return f'<Cycle (id="{self.id}", document_count={len(self.documents)}, completed_tasks={len(completed_tasks)}, incomplete_tasks={len(incomplete_tasks)})>'

    @classmethod
    def load(cls, project, cycle_id):
        cycle_data = project._api.get(f'projects/{project.id}/cycles/{cycle_id}')
        return cls(project, cycle_data)

    @staticmethod
    def generate_id():
        return str(uuid4())[:8]

    @property
    def labels(self):
        labels = []
        # TODO-3: replace with
        #         return [label for task in self.tasks for label in task.labels]
        #         ?
        for task in self.tasks:
            for label in task.labels:
                labels.append(label)
        return labels

    def deserialize_tasks(self, tasks):
        for task_response in tasks:
            task = Task.deserialize(task_response)
            self.tasks.append(task)
            self.documents.append(task.document)

    def to_df(self):
        df = pd.DataFrame(columns=['document_id', 'label', 'annotations'])
        for label in self.labels:
            df = df.append(label.serialize(format='df', project=self.project), ignore_index=True)
        return df.set_index('document_id')

    def add_label(self, task, label):
        current_task = next(t for t in self.tasks if t.id == task.id)
        current_task.labels.append(label)
        current_task.is_complete = True

        data = {
            'cycle_id': str(self.id),
            'task_id': str(current_task.id),
            'label': label.serialize(format='api')
        }
        self.project._api.post(f'projects/{self.project.id}/labels', data)

    def _add_document(self, external_id, document):
        metadata_keys = set(document.keys())
        metadata_keys.remove('contents')

        new_doc = Document(None, external_id, document['contents'], { key: document[key] for key in metadata_keys })
        self.documents.append(new_doc)

    def add_documents(self, from_df=None):
        if from_df is not None:
            for index, doc in from_df.iterrows():
                self._add_document(index, doc)
        else:
            raise Exception('You must provide a collection of documents to add to the cycle.')

    def serialize(self):
        return { 'id': self.id }

    def save(self):
        if self._is_new:
            cycle = self.project._api.post(f'projects/{self.project.id}/cycles', self.serialize())

        for document_batch in chunker(self.documents, MAX_FILE_UPLOAD_BATCH_SIZE):
            document_data = {
                'documents': [doc.serialize() for doc in document_batch]
            }
            data = self.project._api.post('documents/upload', document_data)

            # TODO-3: refactor this, this is inefficient and mutates self.documents
            for doc in document_batch:
                for created in data['documents']:
                    if doc.external_id == created['externalId']:
                        doc.id = created['id']

        created_document_ids = [doc.id for doc in self.documents]
        cycle = self.project._api.post(f'projects/{self.project.id}/cycles/{self.id}/tasks', { 'document_ids': created_document_ids })

        # for now, we just overwrite everything
        self.id = cycle['id']
        self.tasks = []
        self.documents = []
        self.deserialize_tasks(cycle['tasks'])
        self._is_new = False

    def label(self):
        labeler = Labeler(self.project, self)
        display(labeler)

    def next_task(self, skipped_tasks=[]):
        for index, task in enumerate(self.tasks):
            if not task.is_complete and task.id not in skipped_tasks:
                return [task, index]
        raise IndexError('No incomplete tasks found.')



from .project import Project
from .api import Api

class Markers:
    def __init__(self, api_key: str = None, api_url: str = 'https://app.markers.ai'):
        """ api_key should be associated with both project AND user """
        self._api = Api(api_key, api_url)

    def load_project(self, project_id):
        project_data = self._api.get(f'projects/{project_id}')
        project = Project(self._api, project_id, project_data)
        return project



import ipywidgets as widgets # type: ignore
from traitlets import Unicode, Dict, List # type: ignore

from .label import Label

def noop(*args): pass

class MissingCycle(Exception): pass

@widgets.register
class Labeler(widgets.DOMWidget):
    """A data labeling widget to be used with Markers.ai."""
    _view_name = Unicode('AnnotationView').tag(sync=True)
    _model_name = Unicode('AnnotationModel').tag(sync=True)
    _view_module = Unicode('markers').tag(sync=True)
    _model_module = Unicode('markers').tag(sync=True)
    _view_module_version = Unicode('^0.1.0').tag(sync=True)
    _model_module_version = Unicode('^0.1.0').tag(sync=True)

    document = Dict({}).tag(sync=True)
    fragment_classes = List([]).tag(sync=True)
    document_classes = List([]).tag(sync=True)
    metadata = Dict({}).tag(sync=True)

    def __init__(self, project, cycle, **kwargs):
        if not cycle:
            raise MissingCycle('Labeler cannot be created without a cycle.')

        self.cycle = cycle
        super(Labeler, self).__init__(**kwargs)

        self.project = project
        self.document_classes = project.document_classes
        self.fragment_classes = project.fragment_classes
        self.skipped_tasks = []
        self.current_task = None

        self.on_msg(self._handle_message)

        self._load_next_document()

    def _handle_message(self, _, content, buffers):
        handlers = {
            'close_widget': self._handle_close,
            'save_label': self._handle_save_label,
            'skip_document': self._handle_skip_document
        }
        handler = handlers.get(content['event'], noop)
        handler(content.get('data', None))

    def _handle_close(self, *args):
        self.close()

    def _handle_save_label(self, data):
        new_label = Label.create_from_labeler(
            self.document,
            data['label']['name'],
            data['label']['id'],
            self.project,
            data['annotations']
        )
        self.cycle.add_label(self.current_task, new_label)

        self._load_next_document()

    def _handle_skip_document(self, *args):
        self.skipped_tasks.append(self.current_task.id)
        self._load_next_document()

    def _load_next_document(self):
        try:
            [self.current_task, task_index] = self.cycle.next_task(skipped_tasks=self.skipped_tasks)
            self.document = self.current_task.document.serialize()

            self.metadata = {
                'task_progress': task_index + 1,
                'task_count': len(self.cycle.tasks)
            }
        except IndexError as e:
            self.metadata = {
                'status': 'complete'
            }



from ._version import version_info, __version__

from .markers import *
from .project import *
from .labeler import *

def _jupyter_nbextension_paths():
    return [{
        'section': 'notebook',
        'src': 'static',
        'dest': 'markers',
        'require': 'markers/extension'
    }]



import requests, urllib.parse

class Api:
    def __init__(self, api_key, api_url):
        self.api_key = api_key
        self.api_url = api_url

    def get(self, uri):
        params = { 'key': self.api_key }
        uri = '/'.join([self.api_url, 'api', uri])

        res = requests.get(uri, params=params)
        res.raise_for_status()

        data = res.json()
        return data

    def post(self, uri, data):
        params = { 'key': self.api_key }
        uri = '/'.join([self.api_url, 'api', uri])

        res = requests.post(uri, params=params, json=data)
        res.raise_for_status()

        data = res.json()
        return data



# NOTE: reference graphql request in python
# import requests
# url = 'http://localhost:5001/graphql'
# query = """
# query {
#   projectById (id: "a509adcc-5d79-11e9-b5ae-d3399d6dcf54") {
#     id
#     name
#     cycles (filter: { id: { equalTo: "305195d5"}}) {
#       nodes {
#         id
#         tasks {
#           nodes {
#             id
#             document {
#               id
#               externalId
#               contents
#               metadata
#             }
#           }
#         }
#       }
#     }
#   }
# }
# """
# json = { 'query' : query }
# params = { 'api_key': '51af1c9d-8b84-469e-a54e-f754ca86ecbc' }

# res = requests.post(url=url, params=params, json=json)
# data = res.json()
# print(data)



from IPython.display import display, HTML # type: ignore
from typing import List, Optional, Dict
from uuid import UUID
import pandas as pd

from .cycle import Cycle
from .label import Label


class LabelBatch:
    def __init__(self, project, labels):
        self.project: Project = project
        self.labels: List[Label] = labels

    def __len__(self):
        return len(self.labels)

    @classmethod
    def deserialize(cls, project, response):
        labels = [Label.deserialize(label) for label in response]
        return cls(project, labels)

    def to_df(self):
        df = pd.DataFrame(columns=['document_id', 'label', 'annotations'])
        for label in self.labels:
            df = df.append(label.serialize(format='df', project=self.project), ignore_index=True)
        return df.set_index('document_id')


class Project:
    def __init__(self, api, project_id: UUID, project_data=None):
        self._api = api
        self.id: UUID = project_id
        self._labels: LabelBatch = None
        self.metadata = project_data['metadata'] # TODO-3: obviate need for metadata counts with methods that count the actual relations, eg, so that creating new cycle programatically is reflected in counts
        self.cycle_ids = [cycle['id'] for cycle in project_data['cycles']]

        if project_data:
            self.document_classes: List[Dict] = project_data['documentClasses']
            self.fragment_classes: List[Dict] = project_data['fragmentClasses']
        else:
            self.document_classes: List[Dict] = []
            self.fragment_classes: List[Dict] = []

    def __repr__(self):
        return f'<Project (id="{self.id}", label_count={self.metadata["labelCount"]}, document_count={self.metadata["documentCount"]}, cycle_count={len(self.cycle_ids)})>'

    def create_cycle(self) -> Cycle:
        return Cycle(self)

    def load_cycle(self, cycle_id):
        return Cycle.load(self, cycle_id)

    def label(self, df = None, cycle_id: str = None) -> Cycle:
        # TODO-3: allow passing `index_col` to use instead of df index

        if cycle_id:
            cycle = self.load_cycle(cycle_id)
        elif len(self.cycle_ids) == 1 and self.metadata['documentCount'] == 0:
            cycle = self.load_cycle(self.cycle_ids[0])
        else:
            cycle = Cycle(self)
            self.cycle_ids.append(cycle.id)

        if df is not None:
            cycle.add_documents(from_df=df)
            self.metadata['documentCount'] += len(df) # update project so that we don't continue to think count = 0
            cycle.save()

        cycle.label()

        return cycle

    @property
    def labels(self):
        if not self._labels:
            response = self._api.get(f'projects/{self.id}/labels')
            self._labels = LabelBatch.deserialize(self, response['labels'])
        return self._labels

    def open_app(self):
        s = f'<script type="text/javascript">window.open("{self._api.api_url}/projects/{self.id}")</script>'
        display(HTML(s))